@@ -1,6 +1,6 @@
 (*****************************************************************************)
 (*  libCUDF - CUDF (Common Upgrade Description Format) manipulation library  *)
-(*  Copyright (C) 2009-2011  Stefano Zacchiroli <zack@pps.jussieu.fr>        *)
+(*  Copyright (C) 2009-2012  Stefano Zacchiroli <zack@upsilon.cc>            *)
 (*                                                                           *)
 (*  This library is free software: you can redistribute it and/or modify     *)
 (*  it under the terms of the GNU Lesser General Public License as           *)
@@ -110,6 +110,18 @@ let parse_pkgs_test = parse_test ~parse_fun:parse_pkgs_wrapper
 let load_cudf_test = parse_test ~parse_fun:load_cudf_wrapper
 let load_univ_test = parse_test ~parse_fun:load_pkgs_wrapper
 
+(** {!OUnit.assert_equal} proxy for {!Cudf_types.typed_value} values *)
+let assert_equal_tyval ?cmp ?pp_diff ?msg expected real =
+  let printer = Cudf_types_pp.string_of_value in
+  OUnit.assert_equal ?cmp ~printer ?pp_diff ?msg expected real
+
+let assert_equal_string ?cmp ?pp_diff ?msg expected real =
+  let printer s = s in
+  OUnit.assert_equal ?cmp ~printer ?pp_diff ?msg expected real
+
+let string_of_int_list l =
+  sprintf "[%s]" (String.concat ";" (List.map string_of_int l))
+
 (** {5 Test builders} *)
 
 let good_parse ~parse_fun name = TestCase (fun _ ->
@@ -166,7 +178,7 @@ let bad_pkgs_parse_suite =
 
 let value_parse_suite =
   let value_parse_ok (desc, typ, s, v) = desc >: TestCase (fun _ ->
-    assert_equal (Cudf_types_pp.parse_value typ s) v) in
+    assert_equal_tyval (Cudf_types_pp.parse_value typ s) v) in
   let value_parse_ko (desc, typ, s) = desc >: TestCase (fun _ ->
     assert_raises'
       ~cmp:(fun e1 e2 ->
@@ -264,11 +276,74 @@ let value_parse_suite =
     ] ;
   ]
 
+let property_access_suite =
+  let pre, univ, req = load_cudf_test "legacy" in
+  let pre = Option.get pre in
+  let pkg = lookup_package univ ("gasoline-engine", 1) in
+  let pre_ty (p, v) =
+    p >:: (fun _ -> assert_equal_tyval v (lookup_typed_preamble_property pre p)) in
+  let pkg_ty (p, v) =
+    p >:: (fun _ -> assert_equal_tyval v (lookup_typed_package_property pkg p)) in
+  let req_ty (p, v) =
+    p >:: (fun _ -> assert_equal_tyval v (lookup_typed_request_property req p)) in
+  let pre_raw (p, v) =
+    p >:: (fun _ -> assert_equal_string v (lookup_preamble_property pre p)) in
+  let pkg_raw (p, v) =
+    p >:: (fun _ -> assert_equal_string v (lookup_package_property pkg p)) in
+  let req_raw (p, v) =
+    p >:: (fun _ -> assert_equal_string v (lookup_request_property req p)) in
+  "property access" >::: [
+    "raw" >::: [
+      "preamble" >::: List.map pre_raw [
+	"univ-checksum", "8c6d8b4d0cf7027cd523ad095d6408b4901ac31c";
+	"status-checksum", "6936ce910eb716ad97190393f80c14ab04d95b3d";
+	"req-checksum", "17259225eaf63642f9ab99a627b9857a5b27c5f7";
+      ] ;
+      "package" >::: List.map pkg_raw [
+	"package", "gasoline-engine";
+	"version", "1";
+	"depends", "turbo";
+	"provides", "engine";
+	"conflicts", "engine , gasoline-engine";
+	"installed", "true";
+      ] ;
+      "request" >::: List.map req_raw [
+	"request",
+	"http://www.example.org/8f46e388-042f-415e-8aab-df4eeb974444.dudf";
+	"install", "bicycle , electric-engine = 1";
+	"upgrade", "door , wheel > 2";
+      ] ;
+    ] ;
+    "typed" >::: [
+      "preamble" >::: List.map pre_ty [
+	"univ-checksum", (`String "8c6d8b4d0cf7027cd523ad095d6408b4901ac31c");
+	"status-checksum", (`String "6936ce910eb716ad97190393f80c14ab04d95b3d");
+	"req-checksum", (`String "17259225eaf63642f9ab99a627b9857a5b27c5f7");
+      ] ;
+      "package" >::: List.map pkg_ty [
+	"package", (`Pkgname "gasoline-engine");
+	"version", (`Posint 1);
+	"depends", (`Vpkgformula [["turbo", None]]);
+	"provides", (`Veqpkglist ["engine", None]);
+	"conflicts", (`Vpkglist ["engine", None; "gasoline-engine", None]);
+	"installed", (`Bool true);
+      ] ;
+      "request" >::: List.map req_ty [
+	"request",
+	(`String "http://www.example.org/8f46e388-042f-415e-8aab-df4eeb974444.dudf");
+	"install", (`Vpkglist ["bicycle", None;
+			       "electric-engine", Some (`Eq, 1)]);
+	"upgrade", (`Vpkglist ["door", None;
+			       "wheel", Some (`Gt, 2)]);
+      ] ;
+    ]
+  ]
+
 let value_pp_suite =
   let value_pp_ok (desc, v, s) = desc >: TestCase (fun _ ->
-    assert_equal (Cudf_types_pp.string_of_value v) s) in
+    assert_equal_string (Cudf_types_pp.string_of_value v) s) in
   let decl_pp_ok (desc, v, s) = desc >: TestCase (fun _ ->
-    assert_equal (Cudf_types_pp.string_of_typedecl v) s) in
+    assert_equal_string (Cudf_types_pp.string_of_typedecl v) s) in
   "value pretty printing" >::: [
     "good value" >::: List.map value_pp_ok [
     ] ;
@@ -335,40 +410,42 @@ let misc_parse_suite =
   "misc parsing" >::: [
     "qstring" >::: [
       "base" >:: (fun () ->
-        assert_equal (Cudf_types_pp.parse_qstring "\"foo\"") "foo") ;
+        assert_equal_string (Cudf_types_pp.parse_qstring "\"foo\"") "foo") ;
       "escape \"" >:: (fun () ->
-        assert_equal (Cudf_types_pp.parse_qstring "\"fo\\\"o\"") "fo\"o") ;
+        assert_equal_string (Cudf_types_pp.parse_qstring "\"fo\\\"o\"") "fo\"o") ;
       "escape \\" >:: (fun () ->
-        assert_equal (Cudf_types_pp.parse_qstring "\"fo\\\\o\"") "fo\\o") ;
+        assert_equal_string (Cudf_types_pp.parse_qstring "\"fo\\\\o\"") "fo\\o") ;
       "dangling \"" >:: (fun () -> "unexpected parse success" @?
         (try ignore (Cudf_types_pp.parse_qstring "\"fo\"o\"") ; false
 	 with _ -> true)) ;
       "typename ident" >:: (fun () ->
-        assert_equal (Cudf_types_pp.parse_type "ident") `Ident) ;
+        assert_equal ~printer:Cudf_types_pp.string_of_type
+	  (Cudf_types_pp.parse_type "ident") `Ident) ;
     ];
     "pkg comparison" >::: [
       "=%" >:: (fun () ->
         let pkg1 = { default_package with installed = true } in
         let pkg2 = { default_package with installed = false } in
-        assert_equal (pkg1 =% pkg2) true) ;
+	assert_equal ~cmp:(=%) pkg1 pkg2);
       "<%" >:: (fun () ->
         let pkg1 = { default_package with installed = true } in
         let pkg2 = { default_package with installed = false } in
 	let l1 = [ pkg1 ; pkg2 ] in
 	let l2 = [ pkg2 ; pkg1 ] in
-	let rec pkgs_eq = function
+	let rec pkgs_eq pkgs1 pkgs2 =
+	  match pkgs1, pkgs2 with
 	  | [], [] -> true
-	  | p1::t1, p2::t2 -> (p1 =% p2) && pkgs_eq (t1,t2)
+	  | p1::t1, p2::t2 -> (p1 =% p2) && pkgs_eq t1 t2
 	  | _ -> assert false in
-        assert_equal
-	  (pkgs_eq ((List.sort ~cmp:(<%) l1), (List.sort ~cmp:(<%) l2)))
-	  true) ;
+        assert_equal ~cmp:pkgs_eq
+	  (List.sort ~cmp:(<%) l1)
+	  (List.sort ~cmp:(<%) l2));
     ];
   ]
 
 let or_dep =
   "disjunctive dependencies" >:: (fun () ->
-    assert_equal
+    assert_equal ~printer:Cudf_types_pp.string_of_vpkgformula
       (lookup_package (load_univ_test "or-dep") ("electric-engine", 1)).depends
       [["solar-collector", None; "huge-battery", None]])
 
@@ -388,14 +465,24 @@ let status_filtering =
         (fun { installed = i } -> i)
         (get_packages (status univ)))
 
+let status_sizes =
+  "status projection size" >:: (fun _ ->
+    let _, univ, _ = load_cudf_test "legacy" in
+    let status = status univ in
+    let assert_equal' x y = assert_equal ~printer:string_of_int x y in
+    assert_equal' 6 (universe_size status);
+    assert_equal' 6 (installed_size status);
+    assert_equal' (installed_size univ) (installed_size status))
+
 let inst_version_lookup =
   "lookup installed versions" >:: (fun () ->
     let univ = load_univ_test "multi-versions" in
     let versions pkg = List.map (fun p -> p.version) (get_installed univ pkg) in
-      assert_equal (List.sort (versions "gasoline-engine")) [1; 2];
-      assert_equal (versions "battery") [3];
-      assert_equal (versions "not-installed") [];
-      assert_equal (versions "not-existent") [])
+      assert_equal ~printer:string_of_int_list
+	(List.sort (versions "gasoline-engine")) [1; 2];
+      assert_equal ~printer:string_of_int_list (versions "battery") [3];
+      assert_equal ~printer:string_of_int_list (versions "not-installed") [];
+      assert_equal ~printer:string_of_int_list (versions "not-existent") [])
 
 let mem_installed =
   "check whether an installation satisfy a package constraint" >:: (fun () ->
@@ -461,9 +548,11 @@ let univ_sizes =
   let univ = lazy (let _, univ, _ = load_cudf_test "legacy" in univ) in
     "check universe size measuring" >::: [
       "total size" >::
-	(fun () -> assert_equal (universe_size (Lazy.force univ)) 20);
+	(fun () -> assert_equal ~printer:string_of_int
+	  (universe_size (Lazy.force univ)) 20);
       "installed size" >::
-	(fun () -> assert_equal (installed_size (Lazy.force univ)) 6);
+	(fun () -> assert_equal ~printer:string_of_int
+	  (installed_size (Lazy.force univ)) 6);
     ]
 
 let default_value =
@@ -473,7 +562,7 @@ let default_value =
 	(fun () ->
 	  let car = lookup_package (Lazy.force univ) ("car", 1) in
 	  let bugs = List.assoc "bugs" car.pkg_extra in
-	  assert_equal bugs (`Int 0))
+	  assert_equal_tyval (`Int 0) bugs)
     ]
 
 let good_solution_suite = "good solutions" >:::
@@ -494,11 +583,35 @@ let inconsistency_suite = "inconsistent universes" >:::
        not (fst (Cudf_checker.is_consistent (load_univ_test u)))))
     inconsistent_univs
 
+let typedecl_lookup = "type declaration lookup" >:::
+  let pre, _univ, _req = load_cudf_test "legacy" in
+  let pre = Option.get pre in
+  let test_typedecl (prop, typedecl1) =
+    prop >:: (fun _ ->
+      assert_equal ~printer:Std.dump (* TODO use/write a better printer *)
+	typedecl1
+	(lookup_package_typedecl ~extra:pre.property prop)) in
+  List.map
+    test_typedecl
+    [ "package", `Pkgname None;
+      "version", `Posint None;
+      "depends", `Vpkgformula (Some []);
+      "conflicts", `Vpkglist (Some []);
+      "provides", `Veqpkglist (Some []);
+      "installed", `Bool (Some false);
+      "was-installed", `Bool (Some false);
+      "keep", `Enum (keep_enums, Some "none");
+      "suite", `Enum (["stable"; "testing"; "unstable"], Some "stable");
+      "bugs", `Int (Some 0);
+      "description", `String (Some "no description");
+    ]
+
 (** {6 Test suites} *)
 
 let feature_suite =
   "new feature tests" >::: [
     status_filtering ;
+    status_sizes ;
     inst_version_lookup ;
     mem_installed ;
     mem_package ;
@@ -508,6 +621,7 @@ let feature_suite =
     consistency ;
     univ_sizes ;
     default_value ;
+    typedecl_lookup ;
   ]
 
 (*
@@ -538,6 +652,7 @@ let all =
   "all tests" >::: [
     value_parse_suite ;
     value_pp_suite ;
+    property_access_suite ;
     cudf_pp_suite ;
     misc_parse_suite ;
     good_cudf_parse_suite ;
